#!/usr/bin/python

import BluebagConfig

import logging, glob, re, os
from threading import Thread, Event
from time import sleep

class DongleManager(Thread):
  
  def __init__(self, bluebagConfig):
    Thread.__init__(self);
    self.bluebagConfig_ = bluebagConfig;    
    self.busyDongles_ = {};
    self.numDongles_ = 0;
    self.quitEvent_ = Event();
    self.wakeupEvent_ = Event();
    self.validDongleList_ = Event();
    self.canUpdateList_ = Event();

    self.logger_ = logging.getLogger('DongleManager');
    
    if self.bluebagConfig_.autoDetectDongles_:
      self.logger_.info("Autodetecting dongles.");
      dongleList = self.detectDongles();
      for dongle in dongleList:
        self.addDongle(dongle);
    else:
      self.logger_.info("Registering fixed dongles.");
      for count in range(0,self.bluebagConfig_.numDongles_):
        self.addDongle("hci"+str(count));
    self.logger_.info("Registered "+str(self.numDongles_)+" dongles.");
    self.logger_.info("Dongle manager initialized.");
  
  def detectDongles(self):
    dongleList = [];
    globList = glob.glob("/sys/class/bluetooth/hci*");
    if len(globList) > 0:
      for item in globList:
        dongleList.append(item[item.rfind("hci"):]);
      self.logger_.debug("Dongles correctly detected");
    else:
      self.logger_.error("Error trying to autodetect dongles.");
    return dongleList;

  def addDongle(self, device):
    if device in self.busyDongles_:
      self.logger_.warning("Dongle already registered: " + device);
      return False;

    if device in self.bluebagConfig_.reservedDongles_:
      self.logger_.warning("Dongle reserved: " + device);
      return False;
    
    self.numDongles_ += 1;
    self.busyDongles_[device] = False;
    self.logger_.info("Dongle registered: " + device);
    return True;

  def removeDongle(self, device):
    if device not in self.busyDongles_:
      self.logger_.warning("Dongle not registered: " + device);
      return False;

    self.numDongles_ -= 1;
    del self.busyDongles_[device];
    self.logger_.info("Dongle removed: " + device);
    return True;

  def removeDongles(self):
    self.busyDongles_.clear();
    self.numDongles_ = 0;
    self.logger_.info("Dongle list cleared.");
    return True;
  
  def lockDongle(self, device = None):
    if not self.validDongleList_.isSet():
      self.logger_.debug("Wait locking dongle until dongle list is updated.");
      self.validDongleList_.wait();

    self.canUpdateList_.clear();

    if not device:
      for dev, value in self.busyDongles_.items():
        if value == False:
          device = dev;
          
    if not device:
      self.logger_.error("Dongles busy.");
      self.canUpdateList_.set();
      return "";
    
    if device not in self.busyDongles_:
      self.logger_.error("Dongle not registered: " + device);
      self.canUpdateList_.set();
      return "";
    
    if self.busyDongles_[device] == True:
      self.logger_.error("Dongle already locked: " + device);
      self.canUpdateList_.set();
      return "";

    self.busyDongles_[device] = True;
    self.logger_.info("Dongle locked: " + device);
#    self.logger_.debug("Executing: hciconfig "+device+" reset");
#    os.system("hciconfig "+device+" reset");
    self.canUpdateList_.set();
    return device;

  def unlockDongle(self, device):
    if not self.validDongleList_.isSet():
      self.logger_.debug("Wait locking dongle until dongle list is updated.");
      self.validDongleList_.wait();

    if device not in self.busyDongles_:
      self.logger_.error("Dongle not registered: " + device);
      return False;
    
    if self.busyDongles_[device] == False:
      self.logger_.error("Dongle already unlocked: " + device);
      return False;
    
    self.canUpdateList_.clear();
    self.busyDongles_[device] = False;
    self.canUpdateList_.set();

    self.logger_.info("Dongle unlocked: " + device);
    return True;

  def run(self):
    self.validDongleList_.clear();
    self.canUpdateList_.set();
    while True:
      if self.quitEvent_.isSet():
        self.logger_.debug("Time to leave.");
        break;
      
      if self.bluebagConfig_.autoDetectDongles_:
        if not self.canUpdateList_.isSet():
          self.logger_.debug("Wait updating dongle list while used by some thread.");
          self.canUpdateList_.wait();
        self.validDongleList_.clear();
        ### CHECK FOR DONGLES VALIDITY ###
        self.logger_.debug("Updating dongle list.");
        dongleList = self.detectDongles();
        for dongle in dongleList:
          if dongle not in self.busyDongles_:
            self.addDongle(dongle);
        for dongle in self.busyDongles_.keys():
          if dongle not in dongleList:
            self.removeDongle(dongle);
      self.validDongleList_.set();
      
      sleep(5);

  def quitASAP(self):
    self.logger_.debug("Request the dongle manager to quit ASAP.");
    self.quitEvent_.set();
    self.wakeupEvent_.set();

def main():
  # Setting up logging facilities
  logging.basicConfig(level=logging.DEBUG,
                  format='%(asctime)s %(module)s %(levelname)s %(message)s',
                  filename="./donglemanager.log",
                  filemode='w');
  console = logging.StreamHandler();
  console.setLevel(logging.DEBUG);
  formatter = logging.Formatter("%(asctime)s %(module)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);
  
  # Do some testing
  dongleManager = DongleManager();

  logging.getLogger('DongleManager').debug("Registering dongles...");
  dongleManager.addDongle("hci0");
  dongleManager.addDongle("hci1");
  dongleManager.addDongle("hci2");
  dongleManager.addDongle("hci3");
  dongleManager.addDongle("hci4");
  dongleManager.addDongle("hci0");
  for key in dongleManager.busyDongles_.keys():
    logging.getLogger('DongleManager').debug("Dongle: " + key + " - Busy: " + str(dongleManager.busyDongles_[key]));

  logging.getLogger('DongleManager').debug("Unregistering dongles...");
  dongleManager.removeDongle("hci0");
  dongleManager.removeDongle("hci1");
  dongleManager.removeDongle("hci1");
#  for key in dongleManager.busyDongles_.iterkeys():
  for key in dongleManager.busyDongles_.keys():
    logging.getLogger('DongleManager').debug("Dongle: " + key + " - Busy: " + str(dongleManager.busyDongles_[key]));

  logging.getLogger('DongleManager').debug("Locking specific dongles...");
  dongleManager.lockDongle("hci2");
  dongleManager.lockDongle("hci1");
  dongleManager.lockDongle("hci2");
  dongleManager.lockDongle("hci4");
#  for key in dongleManager.busyDongles_.iterkeys():
  for key in dongleManager.busyDongles_.keys():
    logging.getLogger('DongleManager').debug("Dongle: " + key + " - Busy: " + str(dongleManager.busyDongles_[key]));

  logging.getLogger('DongleManager').debug("Locking availables dongles...");
  dongleManager.lockDongle();
  dongleManager.lockDongle();
#  for key in dongleManager.busyDongles_.iterkeys():
  for key in dongleManager.busyDongles_.keys():
    logging.getLogger('DongleManager').debug("Dongle: " + key + " - Busy: " + str(dongleManager.busyDongles_[key]));

  logging.getLogger('DongleManager').debug("Unlocking specific dongles...");
  dongleManager.unlockDongle("hci2");
#  for key in dongleManager.busyDongles_.iterkeys():
  for key in dongleManager.busyDongles_.keys():
    logging.getLogger('DongleManager').debug("Dongle: " + key + " - Busy: " + str(dongleManager.busyDongles_[key]));

  logging.getLogger('DongleManager').debug("Removing dongles...");
  dongleManager.removeDongles();
#  for key in dongleManager.busyDongles_.iterkeys():
  for key in dongleManager.busyDongles_.keys():
    logging.getLogger('DongleManager').debug("Dongle: " + key + " - Busy: " + str(dongleManager.busyDongles_[key]));

  return;

if __name__ == "__main__":
  main();
