#!/usr/bin/python

import BluebagConfig
from QueueManager import QueueManager;
from DongleManager import DongleManager
from DatabaseManager import DatabaseManager
from threading import Thread, Event
import logging, sys, os, subprocess, signal, traceback
from time import sleep

class HciScanWorker(Thread):

  def __init__(self, dongleManager, queueManager, bluebagConfig):
    Thread.__init__(self);
    self.logger_ = logging.getLogger('HciScanWorker');
    self.bluebagConfig_ = bluebagConfig;
    self.dongleManager_ = dongleManager;
    self.queueManager_ = queueManager;
    self.forcedDongle_ = self.bluebagConfig_.scanReservedDongle_;
    self.dongle_ = "";
    self.scanProcess_ = None;
    self.quitEvent_ = Event();
    self.restart_ = Event();
    self.restart_.set();
    self.paused_ = True;
    self.logger_.info("HciScanWorker initialized.");

  def run(self):
    while self.dongle_ == "":
      if self.forcedDongle_ != "":
        self.logger_.debug("Locking dongle: "+self.forcedDongle_);
        self.dongle_ = self.dongleManager_.lockDongle(self.forcedDongle_);
      else:  
        self.logger_.debug("Locking dongle.");
        self.dongle_ = self.dongleManager_.lockDongle();
      if self.dongle_ == "": sleep(1);
    stdin = None;
    stdout = None;
    stderr = None;
    while not self.quitEvent_.isSet():
      try:
        if not self.restart_.isSet():
          self.paused_ = True;
          self.restart_.wait();
          self.paused_ = False;
          if self.quitEvent_.isSet(): break;
        self.paused_ = False;
        self.logger_.debug("Executing: "+self.bluebagConfig_.hcitoolPath_+" -i "+self.dongle_+" scan --flush");

        sleep(5);

        self.scanProcess_ = subprocess.Popen(self.bluebagConfig_.hcitoolPath_+" -i "+self.dongle_+" scan --flush", shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True);
        stdin, stdout, stderr = self.scanProcess_.stdin, self.scanProcess_.stdout, self.scanProcess_.stderr;
        
        stdout.readline();
        output = "";
#        for line in stdout.readlines():
#          output = output + str.strip(line) + "\n";
#        output = str.strip(output);

        numDevFound = 0;
        for line in stdout:
          numDevFound += 1;
          line = str.strip(line);
          output = output + line + "\n";
          device = line[:17];
          self.logger_.debug("Device: "+device);
          if self.queueManager_: self.queueManager_.signalNewDevices([device]);
        self.logger_.info("Found "+str(numDevFound)+" devices.");
##        output = str.strip(output);
        
        err = "";
        for line in stderr.readlines():
          err = err + str.strip(line) + "\n";
        err = str.strip(err);
  
        if err != "":
          self.logger_.warning("Error while scanning using device "+self.dongle_+" : "+err);
##        else:
##          self.logger_.debug("Scanning result: "+output);
##          # Build device list
##          devices = [];
##          for line in output.splitlines():
##            devices.append(line[:17]);
##            self.logger_.debug("Found device: "+line[:17]);
##          self.logger_.info("Found "+str(len(devices))+" devices.");
##          # Signal new devices discovered to the queue manager
##          if (len(devices) > 0) and self.queueManager_: self.queueManager_.signalNewDevices(devices);

      except:
        self.logger_.warning("Caught exception in scan worker, trying to ignore it: "+traceback.format_exc());

      if stdin: stdin.close();
      stdin = None;
      if stdout: stdout.close();
      stdout = None;
      if stderr: stderr.close();
      stderr = None;

      if not self.quitEvent_.isSet(): sleep(self.bluebagConfig_.scanDelay_);
      
    self.logger_.debug("Releasing the dongle.");
    self.dongleManager_.unlockDongle(self.dongle_);
    self.cleanup();
    self.paused_ = True;
    self.restart_.clear();

  def stopScanning(self):
    self.logger_.debug("Forcing scanning to stop immediately.");
    if self.scanProcess_:
      try:
        if self.scanProcess_.poll() == None:
          os.kill(self.scanProcess_.pid, signal.SIGKILL);
          os.waitpid(self.scanProcess_.pid,0);
      except:
        self.logger_.debug("Exception killing scan worker: "+traceback.format_exc());
    self.scanProcess_ = None;
    
  def quitASAP(self):
    self.logger_.debug("Request the worker to quit ASAP.");
    # Kill external processes and perform specific cleanup
    # in order to be able to quit ASAP
    self.quitEvent_.set();

  def cleanup(self):
    # Insert here cleanup to be done right before stopping the thread
    self.stopScanning();
    self.logger_.debug("Releasing the dongle.");
    self.dongleManager_.unlockDongle(self.dongle_);

def main():
  # Setting up logging facilities
  logging.basicConfig(level=logging.DEBUG,
                  format='%(asctime)s %(module)s %(levelname)s %(message)s',
                  filename="./hciscanworker.log",
                  filemode='w');
  console = logging.StreamHandler();
  console.setLevel(logging.DEBUG);
  formatter = logging.Formatter("%(asctime)s %(module)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);
  
  # Do some testing
  dongleManager = DongleManager();
  #dongleManager.addDongle("hci0");
  hciScanWorker = HciScanWorker(dongleManager, None);
  try:
    hciScanWorker.start();
    sleep(40);
#    for count in range (10):
#      print ("sleeping")
#      sleep(20);
#      print ("pause")
#      hciScanWorker.restart_.clear();
#      while not hciScanWorker.paused_:
#        sleep(1);
#        print "waiting...\n";
#      sleep(5);
#      print ("restart")
#      hciScanWorker.restart_.set();
    hciScanWorker.quitASAP();
    hciScanWorker.join();
  except:
    exctype, excvalue = sys.exc_info()[:2];
    logging.getLogger('').fatal("?!??!: " + str(excvalue));
    hciScanWorker.quitASAP();
    hciScanWorker.join();

if __name__ == "__main__":
  main();
